import collections
import itertools


def data(settings) -> str:
    """
    Impute name of field containing base64 encoded data.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        "name_of_data_field"
    """
    return '"' + settings["input"]["name"] + '"'


def fields(settings) -> str:
    """
    Impute name of fields (if any) specifying tensor shape during request.

    Fields can be empty if tensor shape is known beforehand
    (STATIC macro defined, see `header.static` function).

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        "field1", "field2", ..., "fieldN"
    """
    return ", ".join(
        [
            '"' + field + '"'
            for field in settings["input"]["shape"]
            if isinstance(field, str)
        ]
    )


def data_type(settings) -> str:
    type_mapping = {
        "base64": "",
        "byte": "uint8_t",
        "char": "int8_t",
        "short": "int16_t",
        "int": "int32_t",
        "long": "int64_t",
        "float": "float",
        "double": "double",
    }
    return type_mapping[settings["input"]["type"]]


def data_func(settings) -> str:
    type_mapping = {
        "base64": "",
        "byte": "Integer",
        "char": "Integer",
        "short": "Integer",
        "int": "Integer",
        "long": "Int64",
        "float": "Double",
        "double": "Double",
    }
    return "As" + type_mapping[settings["input"]["type"]]


def normalize(settings, key: str) -> str:
    """
    Impute normalization values if any.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict
    key : str
        Name of YAML settings field (either means or stddevs) to be imputed

    Returns
    -------
    str:
        "" or "value1, value2, value3"
    """
    if settings["normalize"] is None:
        return ""
    return ", ".join(map(str, settings["normalize"][key]))


def torch_data_type(settings):
    type_mapping = {
        "base64": "torch::kUInt8",
        "byte": "torch::kUInt8",
        "char": "torch::kInt8",
        "short": "torch::kInt16",
        "int": "torch::kInt32",
        "long": "torch::kInt64",
        "float": "torch::kFloat32",
        "double": "torch::kFloat64",
    }
    return type_mapping[settings["input"]["type"]]


def inputs(settings) -> str:
    """
    Impute input shapes.

    Shapes may be name of fields passed during request (dynamic input shape)
    or integers (static input shape) or mix of both.

    If field is a name (string), it will be transformed to `json_view.GetInteger(name)`.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        String like "1, 3, json_view.GetInteger("width"), json_view.GetInteger("height")"
    """
    return ", ".join(
        str(elem)
        if isinstance(elem, int)
        else 'json_view.GetInteger("{}")'.format(elem)
        for elem in settings["input"]["shape"]
    )


def aws_to_torch(settings, key: str) -> str:
    """
    Impute libtorch specific type from user provided "human-readable" form.

    See `type_mapping` in source code for exact mapping.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        String specifying type, e.g. "torch::kFloat16"
    """
    type_mapping = {
        "bool": "torch::kInt8",
        "int": "torch::kInt32",
        "long": "torch::kInt64",
        "double": "torch::kFloat64",
    }

    if settings["return"][key] is None:
        return ""

    return type_mapping[settings["return"][key]["type"].lower()]


def torch_approximation(settings, key: str) -> str:
    """
    PyTorch has no `bool` type in libtorch hence we approximate one.

    Each item will be later static_casted to appropriate type if needed,
    which is essentially a no-op for for already correct types.

    Usually only `bool` will be casted (eventually other "hard-types" if
    the architecture is specific).

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        String specifying type, e.g. "int8_t"
    """
    type_mapping = {
        "bool": "int8_t",
        "int": "int32_t",
        "long": "int64_t",
        "double": "double",
    }

    if settings["return"][key] is None:
        return ""

    return type_mapping[settings["return"][key]["type"].lower()]


def operations_and_arguments(settings):
    """
    If return->result specified get names of operations to apply on output tensor.

    Merges return->result->operations and return->result->arguments into
    single string to input.

    return ->result->operations is required.

    Names of operations or arguments isn't verified and it may result in compilation
    error if user specifies unavailable tensor function.

    Current weak point in design, check if it can be improved and "safer".

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        string representation of number, e.g. "255.0"

    """

    def _add_namespace(operation):
        return "torch::{}".format(operation)

    def _operation_with_arguments(operation, *values):
        return "{}({})".format(
            _add_namespace(operation),
            ",".join(map(str, [value for value in values if value])),
        )

    def _no_arguments_multiple_operations(operations):
        output = _operation_with_arguments(operations[0], "output")
        for operation in operations[1:]:
            output = _operation_with_arguments(operation, output)
        return output

    def _wrap_in_list(value):
        if not isinstance(value, list):
            return [value]
        return value

    if settings["return"]["result"] is None:
        return ""

    if "code" in settings["return"]["result"]:
        return settings["return"]["result"]["code"]

    operations = settings["return"]["result"]["operations"]
    arguments = settings["return"]["result"]["arguments"]
    if arguments is None:
        if isinstance(operations, str):
            return "{}(output)".format(_add_namespace(operations))
        return _no_arguments_multiple_operations(operations)

    operations, arguments = _wrap_in_list(operations), _wrap_in_list(arguments)
    output = _operation_with_arguments(operations[0], "output", arguments[0])
    for operation, argument in itertools.zip_longest(operations[1:], arguments[1:]):
        output = _operation_with_arguments(operation, output, argument)
    return output


def aws_function(settings, key: str, array: bool) -> str:
    """
    Internal imputation specifying one of AWS SDK functions based on type.

    This function specifies which AWS SDK function will be used to create
    JSONValue (either as single item to return or as part of array to return).

    Looked for in return->output and return->result settings and returns one or both
    depending on which is specified.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict
    key : str
        Name of field to look for in type (either "output" or "result")
    array: bool
        Whether prefix should be tailored to array output (`As`) or item (`With`)

    Returns
    -------
    str:
        "" or "As<type>" or "With<type>" AWS function name

    """

    prefix = "As" if array else "With"
    type_mapping = {
        "int": "Integer",
        "long": "Int64",
        "double": "Double",
    }

    if settings["return"][key] is None:
        return ""
    return prefix + type_mapping[settings["return"][key]["type"].lower()]


def field_if_exists(settings, key: str, name: str) -> str:
    """
    Return value of nested fields if those are specified.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict
    key : str
        Name of field to look for in type (either "output" or "result")
    name: str
        Name of field to look for in one of "output" or "result"

    Returns
    -------
    str:
        "" or value provided in field

    """
    if settings["return"][key] is None:
        return ""
    return settings["return"][key][name]


def model(settings) -> str:
    """
    Return path to model specified by settings.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        "/path/to/model"

    """
    return '"' + settings["model"] + '"'



from . import macro


def static(settings) -> str:
    """
    Return #define STATIC if all inputs fields are integer.

    If specified, no field checks will be performed during request to
    Lambda function.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define STATIC"
    """
    return macro.conditional(
        all(isinstance(x, int) for x in settings["input"]["shape"]), "STATIC"
    )


def grad(settings) -> str:
    """
    Return #define GRAD if grad: True specified.

    If specified, libtorch's grad addition to ATen will be turned on.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define GRAD"
    """
    return macro.conditional(settings["grad"], "GRAD")


def optimize(settings) -> str:
    """
    Return #define GRAD if grad: True specified.

    If specified, libtorch's grad addition to ATen will be turned on.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define GRAD"
    """
    return macro.conditional(settings["optimize"], "OPTIMIZE")


def validate_json(settings) -> str:
    """
    Return #define VALIDATE_JSON if validate_json: True specified.

    If specified, it will be validated if JSON was parsed successfully.
    If not, InvalidJson with appropriate text will be returned.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define VALIDATE_JSON"
    """
    return macro.conditional(settings["validate_json"], "VALIDATE_JSON")


def base64(settings) -> str:
    """
    Return #define BASE64 if input->type is equal to base64.

    If specified, it is assumed data is in base64 string format and will be decoded
    in handler. In this case input_type IS NOT USED and data type will be
    unsigned int 8 upon creation (which can be optionally casted).

    If not, it is assumed data will be a flat array of specified by user via
    input_type.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define BASE64"
    """
    return macro.conditional(settings["input"]["type"] == "base64", "BASE64")


def validate_field(settings) -> str:
    """
    Return #define VALIDATE_FIELD if validate: True specified.

    If specified, data will be checked for correctness
    (whether `data` field exists and whether it's type is string or array).

    If not, InvalidJson with appropriate text will be returned.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define VALIDATE_FIELD"
    """
    return macro.conditional(settings["input"]["validate"], "VALIDATE_FIELD")


def validate_shape(settings) -> str:
    """
    Return #define VALIDATE_SHAPE if validate_inputs: True specified.

    If specified, input fields will be checked (if any exist).
    All of them will be checked for existence and whether their type
    is integer.

    If not, InvalidJson with appropriate text will be returned.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define VALIDATE_SHAPE"
    """
    return macro.conditional(settings["input"]["validate_shape"], "VALIDATE_SHAPE")


def normalize(settings) -> str:
    """
    Return #define NORMALIZE if normalize: True specified.

    Will normalize image-like tensor across channels using
    torch::data::transforms::Normalize<>.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define NORMALIZE"
    """
    return macro.conditional(settings["normalize"], "NORMALIZE")


def cast(settings) -> str:
    """
    Impute libtorch specific type from user provided "human-readable" form.

    See `type_mapping` in source code for exact mapping.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        String specifying type, e.g. "torch::kFloat16"
    """
    type_mapping = {
        "byte": "torch::kUInt8",
        "char": "torch::kInt8",
        "short": "torch::kInt16",
        "int": "torch::kInt32",
        "long": "torch::kInt64",
        "half": "torch::kFloat16",
        "float": "torch::kFloat32",
        "double": "torch::kFloat64",
    }

    return macro.key(settings["input"], key="cast", mapping=type_mapping,)


def divide(settings) -> str:
    """
    Impute value by which casted tensor will be divided.

    If user doesn't want to cast tensor, he should simply specify `1`,
    though it won't be used too often.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        string representation of number, e.g. "255.0"

    """
    return macro.key(settings["input"], key="divide")


def return_output(settings):
    """
    Return #define RETURN_OUTPUT if field return->output is specified.

    Will return output from neural network as JSON array of specified
    OUTPUT_TYPE if macro defined.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define RETURN_OUTPUT"
    """
    return macro.conditional(
        settings["return"]["output"] and not settings["return"]["output"]["item"],
        "RETURN_OUTPUT",
    )


def return_output_item(settings):
    """
    Return #define RETURN_OUTPUT if field return->output->item is specified.

    Will return output from neural network as single element of specified
    OUTPUT_TYPE if macro defined.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define RETURN_OUTPUT_ITEM"
    """
    return macro.conditional(
        settings["return"]["output"] and settings["return"]["output"]["item"],
        "RETURN_OUTPUT_ITEM",
    )


def return_result(settings):
    """
    Return #define RETURN_RESULT if field return->result is specified.

    Will return modified by RESULT_OPERATION output from neural network
    as JSON array of specified RESULT_TYPE if macro defined.

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define RETURN_RESULT"
    """
    return macro.conditional(
        settings["return"]["result"] and not settings["return"]["result"]["item"],
        "RETURN_RESULT",
    )


def return_result_item(settings):
    """
    Return #define RETURN_RESULT_ITEM if field return->result->item is specified.

    Will return modified by RESULT_OPERATION output from neural network
    as single item of specified RESULT_TYPE if macro defined

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Either "" or "#define RETURN_RESULT_ITEM"
    """
    return macro.conditional(
        settings["return"]["result"] and settings["return"]["result"]["item"],
        "RETURN_RESULT_ITEM",
    )



import pathlib
import typing


def validate(args) -> None:
    """
    Validate torchscript model is a file. Exit with 1 otherwise

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    """
    source = pathlib.Path(args.source)
    if not source.is_file:
        print("torchlambda:: Error: provided path to torchscript model is not a file!")
        exit(1)


def path(args) -> pathlib.Path:
    """
    Return path where model will be placed upon `.zip` unpacking.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    Returns
    -------
    pathlib.Path

    """
    if args.directory is not None:
        return pathlib.Path(args.directory) / args.source
    return pathlib.Path(args.source)


def compression_level(compression, level):
    """
    Validate compression and compression's level user arguments.

    This function return appropriate compression level for any `zipfile.Zipfile`
    supported values.

    Check: https://docs.python.org/3/library/zipfile.html#zipfile-objects
    to see possible values for each type.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    Returns
    -------
    pathlib.Path

    """

    def _wrong_parameters(minimum, maximum):
        print(
            "--level should be in range [{}, {}] for compression type {}".format(
                minimum, maximum, level
            )
        )
        exit(1)

    if compression == "DEFLATED":
        if not 0 <= level <= 9:
            _wrong_parameters(0, 9)
        return level
    if compression == "BZIP2":
        if not 1 <= level <= 9:
            _wrong_parameters(1, 9)
        return level

    return level



import pathlib
import zipfile

from .. import implementation


def run(args):
    """Entrypoint for `torchlambda layer` subcommand"""
    destination = pathlib.Path(args.destination).absolute()
    with implementation.general.message(
        "packing Torchscript model to {}.".format(destination)
    ):
        with zipfile.ZipFile(
            destination,
            "w",
            compression=getattr(zipfile, "ZIP_{}".format(args.compression)),
            compresslevel=implementation.layer.compression_level(
                args.compression, args.compression_level
            ),
        ) as file:
            implementation.layer.validate(args)
            file.write(
                pathlib.Path(args.source),
                implementation.layer.path(args)
                if args.directory is not None
                else args.source,
            )



import pathlib

from .. import implementation


def run(args):
    """Entrypoint for `torchlambda settings` subcommand"""
    destination = pathlib.Path(args.destination).absolute()
    with implementation.general.message(
        "creating YAML settings at {}.".format(destination)
    ):
        implementation.general.run(
            "cp -r ./templates/settings/torchlambda.yaml {}".format(destination),
            operation="copying YAML source code",
            silent=args.silent,
        )



import importlib

from . import arguments, subcommands
from ._version import __version__


def main() -> None:
    """Entrypoint run after torchlambda command.

    Responsibilities for functionalities are fully transferred to specific
    subcommand.

    See subcommands module for details.

    """

    args = arguments.parser.get()
    module = importlib.import_module(
        ".subcommands." + args.subcommand, package=__package__
    )
    module.run(args)


if __name__ == "__main__":
    main()



import typing


def define(name: str, value: str = "") -> str:
    """
    Return C++ macro-string #define name value

    Used for header defines.

    Parameters
    ----------
    name : str
        Name of macro to be defined
    value : str, optional
        Value of defined macro (if any),

    Returns
    -------
    str:
        "#define name" or "#define name value"
    """
    return "#define {} {}".format(name, value)


def key(dictionary, key: str, mapping: typing.Dict = None) -> str:
    """
    Return C++ macro-string "#define name value" based on provided key.

    Used for header defines based on user-provided setting.

    Parameters
    ----------
    name : str
        Name of macro to be defined
    value : str, optional
        Value of defined macro (if any),

    Returns
    -------
    str:
        "#define name value" or "#define name value"
    """
    value = dictionary.get(key, None)
    if value is not None:
        if mapping is not None:
            value = mapping[value]
        return define(key.upper(), value)
    return ""


def conditional(condition: bool, name: str, value: typing.Any = "") -> str:
    """
    Return C++ macro-string "#define name value" if condition is True.

    Used for header defines based on user-provided setting.

    Parameters
    ----------
    name : str
        Name of macro to be defined
    value : str, optional
        Value of defined macro (if any),

    Returns
    -------
    str:
        "#define name" or "#define name value"
    """
    if condition:
        return define(name, value)
    return ""



import pathlib
import shutil
import sys
import uuid

from . import docker, general


def get_image(args) -> str:
    """
    Get torchlambda deployment image.

    If image specified by --image exists locally it's name will be returned.
    Else if image should be build from source it is built and it's name is returned
    as specified by --image.
    Otherwise pre-built torchlambda:latest image will be used.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    Returns
    -------
    str:
        Name of obtained image
    """

    def _custom_build(args):
        """True if any of the flags specified as non-default."""
        flags = (
            "pytorch",
            "aws",
            "operations",
            "aws_components",
            "docker_build",
            "pytorch_version",
        )
        # If any is not None or empty list
        return any(map(lambda flag: getattr(args, flag), flags))

    image_exists: bool = docker.image_exists(args.image)
    if image_exists:
        print(
            "torchlambda:: Image {} was found locally and will be used.".format(
                args.image
            )
        )
        return args.image
    print("torchlambda:: Image {} was not found locally.".format(args.image))
    if _custom_build(args):
        general.copy_operations(args)
        return docker.build(args)

    print("torchlambda:: Default szymonmaszke/torchlambda:latest image will be used.")
    return "szymonmaszke/torchlambda:latest"


def get_package(args, image):
    """Generate deployment package by running provided image.

    Deployment package will be .zip file containing compiled source code
    and statically linked AWS and Libtorch (with optional user-specified flags).

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.
    image : str
        Name of image used to create container.
    """
    container: str = docker.run(args, image)
    with docker.rm(container):
        destination = pathlib.Path(args.destination).absolute()
        if not destination.is_file():
            docker.cp(
                args, container, "/usr/local/build/torchlambda.zip", destination,
            )
        else:
            print(
                "torchlambda:: Error: path {} exists. Please run the script again with the same --image argument.".format(
                    destination
                )
            )



import pathlib

from .. import implementation


def run(args):
    """Entrypoint for `torchlambda template` subcommand"""
    with implementation.general.message(
        "creating C++ scheme at {}.".format(args.destination)
    ):
        if args.yaml is None:
            implementation.general.run(
                "cp -r ./templates/custom {}".format(
                    pathlib.Path(args.destination).absolute()
                ),
                operation="copying CPP sources",
                silent=args.silent,
            )
        else:
            implementation.template.create_template(args)



import contextlib
import pathlib
import shutil
import sys
import typing
import uuid

from . import general


def check() -> None:
    """Check whether docker is available for usage by torchlambda."""
    if shutil.which("docker") is None:
        print(
            "Docker is required but was not found on $PATH.\n"
            "Please install docker and make it available for torchlambda (see https://docs.docker.com/install/).",
            file=sys.stderr,
        )
        exit(1)


def image_exists(name: str) -> bool:
    """
    Return true if specified image exist locally on host.

    Parameters
    ----------
    name : str
        Name of image to be searched for

    Returns
    -------
    bool
        Whether image exists on localhost

    """
    return not bool(
        general.run(
            "docker inspect --type=image {}".format(name),
            operation="image existence check.",
            silent=True,
            exit_on_failure=False,
        )
    )


def build(args) -> str:
    """
    Build docker image from scratch.

    User can provide various flags to docker and build commands except `-t` which
    is specified by `image` argument and parsed separately.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    Returns
    -------
    str:
        Name of created image
    """

    def _cmake_environment_variables(name: str, values: typing.List[str]) -> str:
        return (
            '--build-arg {}="{}" '.format(
                name, " ".join(["-D{}".format(value) for value in values]),
            )
            if values
            else " "
        )

    def _create_aws_components(components):
        return ["BUILD_ONLY='{}'".format(";".join(components))] if components else []

    def _pytorch_version(version: str):
        return (
            '--build-arg PYTORCH_VERSION="{}" '.format(version)
            if version is not None
            else " "
        )

    command = "docker {} build {} -t {} ".format(
        *general.parse_none(args.docker, args.docker_build, args.image)
    )
    command += _cmake_environment_variables("PYTORCH", args.pytorch)
    command += _pytorch_version(args.pytorch_version)
    command += _cmake_environment_variables(
        "AWS", _create_aws_components(args.aws_components) + args.aws
    )
    command += "."
    general.run(command, operation="custom PyTorch build.", silent=args.silent)

    return args.image


def run(args, image: str) -> str:
    """
    Run docker image and mount user-provided folder with C++ files.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.
    image : str
        Name of image from which container is run

    Returns
    -------
    str:
        Name of created container. Consist of torchlambda prefix and random string
    """

    def _add_components(args):
        return (
            '"' + ";".join(args.aws_components) + '"'
            if args.aws_components
            else '"core"'
        )

    def _compilation(args):
        return '"' + args.compilation + '"' if args.compilation else ""

    container_name = "torchlambda-" + str(uuid.uuid4())
    source_directory = pathlib.Path(args.source).absolute()
    if source_directory.is_dir():
        command = "docker {} run {} -v {}:/usr/local/user_code --name {} {} {} ".format(
            *general.parse_none(
                args.docker,
                args.docker_run,
                source_directory,
                container_name,
                image,
                _add_components(args),
            )
        )

        command += _compilation(args)

        general.run(
            command,
            operation="building inference AWS Lambda package.",
            silent=args.silent,
        )
        return container_name

    print("torchlambda:: Provided source files are not directory, exiting.")
    exit(1)


def cp(args, name: str, source: str, destination: str) -> None:
    """
    Copy source to destination from named container.

    Will verbosely copy file from container, yet no output of `docker` command
    will be displayed.

    Will crash the program if anything goes wrong.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.
    args : dict-like
        User provided arguments
    name : str
        Name of container
    source : str
        Path to source to be copied
    destination : str
        Destination where source will be copied

    """
    general.run(
        "docker container cp {}:{} {}".format(
            name, source, pathlib.Path(destination).absolute()
        ),
        operation="copying {} from docker container {} to {}.".format(
            source, name, destination
        ),
        silent=args.silent,
    )


@contextlib.contextmanager
def rm(name: str) -> None:
    """
    Remove container specified by name

    Container to be removed after deployment was copied from it to localhost.

    Parameters
    ----------
    name : str
        Name of the container to be removed

    """
    try:
        yield
    finally:
        general.run(
            "docker rm {}".format(name),
            operation="removing created docker container named {}.".format(name),
            silent=True,
        )



import argparse

from .subparsers import build, layer, settings, template


def get():
    """Get user provided arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="Lightweight tool to deploy PyTorch models on AWS Lambda.\n"
        "For more information see documentation: https://github.com/szymonmaszke/torchlambda/wiki",
    )

    parser.add_argument(
        "--silent",
        required=False,
        action="store_true",
        help="Will only output torchlambda information (e.g. building PyTorch output won't be displayed).\n"
        "Default: False",
    )

    subparsers = parser.add_subparsers(help="Subcommands:", dest="subcommand")
    settings(subparsers)
    template(subparsers)
    build(subparsers)
    layer(subparsers)

    return parser.parse_args()



import collections

import cerberus


class Validator(cerberus.Validator):
    def _validate_is_shorter(self, other: str, field, value):
        """Test whether one field (arguments) is shorter than other (operations).

        The rule's arguments are validated against this schema:
        {"type": "string"}
        """
        arguments_length = (
            1 if not isinstance(value, collections.Iterable) else len(value)
        )
        operations_length = (
            1 if isinstance(self.document[other], str) else len(self.document[other])
        )
        if arguments_length > operations_length:
            self._error(
                field,
                "More arguments provided in field: {} than in operations!".format(
                    field,
                ),
            )

    def _validate_broadcastable(self, shape_field: str, field, value):
        """Test whether one field (arguments) is shorter than other (operations).

        The rule's arguments are validated against this schema:
        {"type": "string"}
        """
        shapes = self.root_document["input"][shape_field]
        if len(value) != 1 or len(value) != len(shapes):
            self._error(
                field,
                "{} field's shape is not broadcastable to provided input shape: {}".format(
                    field, shapes
                ),
            )


def _is_not_dict(field, value, error):
    if isinstance(value, dict):
        error(field, "field cannot be an instance of dict")


# Fix validator
def get():
    return Validator(
        {
            "grad": {"type": "boolean", "default": False},
            "optimize": {"type": "boolean", "default": False},
            "validate_json": {"type": "boolean", "default": True},
            "model": {"type": "string", "default": "/opt/model.ptc", "empty": False},
            "input": {
                "type": "dict",
                "schema": {
                    "name": {"type": "string", "default": "data", "empty": False},
                    "validate": {"type": "boolean", "default": True},
                    "type": {
                        "type": "string",
                        "allowed": [
                            "base64",
                            "byte",
                            "char",
                            "short",
                            "int",
                            "long",
                            "float",
                            "double",
                        ],
                        "required": True,
                    },
                    "shape": {
                        "type": "list",
                        "schema": {"type": ["string", "integer"]},
                        "required": True,
                        "minlength": 2,
                        "empty": False,
                    },
                    "validate_shape": {"type": "boolean", "default": True},
                    "cast": {
                        "type": "string",
                        "allowed": [
                            "byte",
                            "char",
                            "short",
                            "int",
                            "long",
                            "half",
                            "float",
                            "double",
                        ],
                        "nullable": True,
                        "default": None,
                    },
                    "divide": {"type": "number", "nullable": True, "default": None},
                },
            },
            "normalize": {
                "type": "dict",
                "nullable": True,
                "schema": {
                    "means": {
                        "required": True,
                        "broadcastable": "shape",
                        "type": "list",
                        "schema": {"type": "number"},
                    },
                    "stddevs": {
                        "required": True,
                        "broadcastable": "shape",
                        "type": "list",
                        "schema": {"type": "number"},
                    },
                },
                "default": None,
            },
            # Return can be either result, output or both
            "return": {
                "type": "dict",
                "required": True,
                "schema": {
                    "output": {
                        "type": "dict",
                        "nullable": True,
                        "schema": {
                            # Return only single item, not array
                            "name": {
                                "type": "string",
                                "default": "output",
                                "empty": False,
                            },
                            "type": {
                                "type": "string",
                                "allowed": ["int", "long", "double", "bool"],
                                "required": True,
                            },
                            "item": {"type": "boolean", "default": False},
                        },
                        "default": None,
                    },
                    "result": {
                        "type": "dict",
                        "nullable": True,
                        "schema": {
                            # Return only single item, not array
                            "name": {"type": "string", "default": "result"},
                            "type": {
                                "type": "string",
                                "allowed": ["int", "long", "double", "bool"],
                                "required": True,
                            },
                            "item": {"type": "boolean", "default": False},
                            "operations": {
                                "oneof": [
                                    {
                                        "type": "list",
                                        "schema": {"type": "string", "empty": False},
                                        "empty": False,
                                    },
                                    {"type": "string", "empty": False},
                                ],
                                "required": True,
                            },
                            "arguments": {
                                "is_shorter": "operations",
                                "check_with": _is_not_dict,
                                "empty": False,
                                "dependencies": "operations",
                                "nullable": True,
                                "default": None,
                            },
                        },
                        "default": None,
                    },
                },
            },
        }
    )



import contextlib
import pathlib
import shlex
import shutil
import subprocess
import sys
import typing


def parse_none(*args):
    """Return tuple of arguments excluding the ones being None."""
    return tuple(arg if arg is not None else "" for arg in args)


@contextlib.contextmanager
def message(operation: str):
    """
    Verbose logging of operation to be performed.

    Used only by `utils.run_command`

    Parameters
    ----------
    operation : str
        String describing operation to be run.

    """
    print("torchlambda:: Started {}".format(operation), file=sys.stderr)
    yield
    print("torchlambda:: Finished {}".format(operation), file=sys.stderr)


def run(
    command: str,
    operation: str,
    silent: bool,
    exit_on_failure: bool = True,
    no_stdout: bool = False,
    no_stderr: bool = False,
) -> int:
    """
    Run specific cli command, parse and return it's results.

    Parameters
    ----------
    command : str
        CLI Linux command to be run.
    operation : str
        Name of operation for verbose logging of output.
    exit_on_failure : bool, optional
        Whether program should exit with error if the command didn't return 0.
        Default: `True`

    Returns
    -------
    int
        Value returned by command

    """

    def _set_streams() -> typing.Dict:
        kwargs = {}
        if no_stdout or silent:
            kwargs["stdout"] = subprocess.DEVNULL
        if no_stderr or silent:
            kwargs["stderr"] = subprocess.DEVNULL
        return kwargs

    with message(operation):
        return_value = subprocess.call(
            shlex.split(command),
            cwd=pathlib.Path(__file__).absolute().parent.parent,
            **_set_streams()
        )
        if return_value != 0 and exit_on_failure:
            print(
                "torchlambda:: Error: Failed during {}".format(operation),
                file=sys.stderr,
            )
            exit(1)

        return return_value


@message("copying provided model operations.")
def copy_operations(args) -> None:
    """
    Copy custom operations.yaml of PyTorch model if provided by user.

    Providing custom operations.yaml will trigger docker image build from scratch.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    """
    if args.operations is not None:
        operations_path = pathlib.Path(args.operations).absolute()
        if operations_path.is_file():
            print("torchlambda:: Copying custom model operations...")
            shutil.copyfile(
                pathlib.Path(args.operations).absolute(),
                pathlib.Path(__file__).absolute().parent.parent / "model.yaml",
            )
        else:
            print("torchlambda:: Error, provided operations are not a file!")
            exit(1)



from .. import implementation


def run(args):
    """Entrypoint for `torchlambda build` subcommand"""
    implementation.docker.check()

    with implementation.general.message("deployment."):
        image = implementation.build.get_image(args)
        if not args.no_run:
            implementation.build.get_package(args, image)



import pathlib
import sys
import typing

import yaml

from . import general, utils


@general.message("reading YAML settings.")
def read_settings(args) -> typing.Dict:
    """
    Read user provided YAML settings.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    Returns
    -------
        Dictionary containing settings to create template
    """
    with open(args.yaml, "r") as file:
        try:
            return yaml.safe_load(file)
        except yaml.YAMLError as error:
            print("torchlambda:: Error during user settings parsing:")
            print(error)
            sys.exit(1)


@general.message("validating YAML settings.")
def validate(validator, settings: typing.Dict) -> None:
    """
    Validate user provided YAML settings.

    Uses custom validator, see `utils.template/validator.py` for
    exact validation scheme.

    Parameters
    ----------
    validator : cerberus.Validator
        Cerberus scheme validator
    settings : typing.Dict
        YAML parsed to dict

    """
    if not validator.validate(settings, normalize=True):
        print("torchlambda:: Error during YAML validation:", file=sys.stderr)
        print(yaml.dump(validator.errors), file=sys.stderr)
        sys.exit(1)


@general.message("creating .cpp source from YAML template.")
def create_source(settings: typing.Dict) -> str:
    """
    Create .cpp source code from template and settings

    Values are imputed using `format` to `main.cpp`.
    Each field is either a header (#define NAME <value>, <value> optional)
    or is directly imputed into source code.

    First case and all needed processing is located in `utils.template.header`,
    second case and all needed processing is located in `utils.template.imputation`

    `template/main.cpp` template file uses double curly brackets ({{}}) where normal
    C++ brackets are used in order to be compatible with `str.format`, see
    this: https://stackoverflow.com/questions/5466451/how-can-i-print-literal-curly-brace-characters-in-python-string-and-also-use-fo
    for some info.

    **DESCRIPTION OF HEADER FIELDS**:

        - STATIC - whether all shapes are static (e.g. no
        input dimension is dependent on value passed in request as field).
        True if `input_shape` only has integers (fixed input shape).

        - VALIDATE_INPUTS - whether fields provided in INPUT_SHAPE should be
        checked for correctness (they exist and are of integer type).
        Default: True

        - NO_GRAD - whether PyTorch's gradient should be disabled.
        Usually yes as AWS Lambda is mainly used for inference
        Default: True

        - CAST - to which type should the tensor be casted after creation from
        base64 encoded data (by default it's unsigned char which is rarely useful).
        Default: float

        - NORMALIZE - whether normalize input using NORMALIZE_MEANS and NORMALIZE_STDDEVS
        Used mainly for image inference.
        Default: False

        - DIVIDER - value by which input tensor will be divided after casting
        to CAST.
        Default: 255 (to bring `unsigned char` to `[0,1]` range, useful for image inference)

        - RETURN_OUTPUT - whether to return output as array.
        Exclusive with RETURN_OUTPUT_ITEM

        - RETURN_OUTPUT_ITEM - whether to return output as single item.
        Exclusive with RETURN_OUTPUT

        - RETURN_RESULT - whether to return result as array.
        Exclusive with RETURN_RESULT_ITEM

        - RETURN_RESULT_ITEM - whether to return output as single item.
        Exclusive with RETURN_RESULT

    **DESCRIPTION OF DIRECTLY IMPUTED FIELDS**:

        - DATA - Name of field providing data in request

        - FIELDS - Names (if any) of provided non-static fields

        - NORMALIZE_MEANS - Means to use for normalization (if any)

        - NORMALIZE_STDDEVS - Standard deviations to use for normalization (if any)

        - INPUTS - Input shapes used for reshape of tensor (including batch dimension)

        - IF RETURN_RESULT OR RETURN_RESULT_ITEM DEFINED:

            - OPERATIONS_AND_ARGUMENTS - Operations with arguments (if any) to apply over
            outputted tensor to create result.
            Mix of "operations" and "arguments" fields

        - IF RETURN_OUTPUT OR RETURN_OUTPUT_ITEM DEFINED:

            - AWS_OUTPUT_FUNCTION - Internal AWS SDK JsonValue function like WithInteger.
            Based on RETURN_OUTPUT_TYPE

        - IF RETURN_RESULT OR RETURN_RESULT_ITEM DEFINED:

            - AWS_RESULT_FUNCTION - Internal AWS SDK JsonValue function like WithInteger.
            Based on RETURN_RESULT_TYPE

        - IF RETURN_OUTPUT OR RETURN_OUTPUT_ITEM DEFINED:

            - OUTPUT_NAME - Key of returned JSON
            - OUTPUT_TYPE - Type  of item of returned JSON

        - IF RETURN_RESULT OR RETURN_RESULT_ITEM DEFINED:

            - RESULT_NAME - Key of returned JSON
            - RESULT_TYPE - Type  of item of returned JSON

    Parameters
    ----------
    settings : typing.Dict
        YAML parsed to dict

    Returns
    -------
    str:
        Source represented as string
    """
    cwd = pathlib.Path(__file__).absolute().parent.parent
    with open(cwd / "templates/settings/main.cpp") as file:
        return file.read().format(
            # Top level defines
            STATIC=utils.template.header.static(settings),
            GRAD=utils.template.header.grad(settings),
            OPTIMIZE=utils.template.header.optimize(settings),
            VALIDATE_JSON=utils.template.header.validate_json(settings),
            BASE64=utils.template.header.base64(settings),
            VALIDATE_FIELD=utils.template.header.validate_field(settings),
            VALIDATE_SHAPE=utils.template.header.validate_shape(settings),
            NORMALIZE=utils.template.header.normalize(settings),
            CAST=utils.template.header.cast(settings),
            DIVIDE=utils.template.header.divide(settings),
            RETURN_OUTPUT=utils.template.header.return_output(settings),
            RETURN_OUTPUT_ITEM=utils.template.header.return_output_item(settings),
            RETURN_RESULT=utils.template.header.return_result(settings),
            RETURN_RESULT_ITEM=utils.template.header.return_result_item(settings),
            # Direct insertions
            DATA=utils.template.imputation.data(settings),
            FIELDS=utils.template.imputation.fields(settings),
            DATA_TYPE=utils.template.imputation.data_type(settings),
            DATA_FUNC=utils.template.imputation.data_func(settings),
            NORMALIZE_MEANS=utils.template.imputation.normalize(settings, key="means"),
            NORMALIZE_STDDEVS=utils.template.imputation.normalize(
                settings, key="stddevs"
            ),
            TORCH_DATA_TYPE=utils.template.imputation.torch_data_type(settings),
            INPUTS=utils.template.imputation.inputs(settings),
            OUTPUT_CAST=utils.template.imputation.aws_to_torch(settings, "output"),
            RESULT_CAST=utils.template.imputation.aws_to_torch(settings, "result"),
            OPERATIONS_AND_ARGUMENTS=utils.template.imputation.operations_and_arguments(
                settings
            ),
            AWS_OUTPUT_FUNCTION=utils.template.imputation.aws_function(
                settings, key="output", array=True,
            ),
            AWS_RESULT_FUNCTION=utils.template.imputation.aws_function(
                settings, key="result", array=True,
            ),
            AWS_OUTPUT_ITEM_FUNCTION=utils.template.imputation.aws_function(
                settings, key="output", array=False
            ),
            AWS_RESULT_ITEM_FUNCTION=utils.template.imputation.aws_function(
                settings, key="result", array=False
            ),
            TORCH_OUTPUT_TYPE=utils.template.imputation.torch_approximation(
                settings, key="output"
            ),
            OUTPUT_TYPE=utils.template.imputation.field_if_exists(
                settings, key="output", name="type"
            ),
            TORCH_RESULT_TYPE=utils.template.imputation.torch_approximation(
                settings, key="result"
            ),
            RESULT_TYPE=utils.template.imputation.field_if_exists(
                settings, key="result", name="type"
            ),
            OUTPUT_NAME=utils.template.imputation.field_if_exists(
                settings, key="output", name="name"
            ),
            RESULT_NAME=utils.template.imputation.field_if_exists(
                settings, key="result", name="name"
            ),
            MODEL_PATH=utils.template.imputation.model(settings),
        )


def save(source: str, args) -> None:
    """
    Save created source file to specified destination.

    All needed folders will be created if necessary.

    Parameters
    ----------
    source : str
        String representation of C++ deployment source code
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    """
    destination = pathlib.Path(args.destination).absolute()
    destination.mkdir(parents=True, exist_ok=True)
    with open(destination / "main.cpp", "w") as file:
        file.write(source)


def create_template(args) -> None:
    """
    Create and save template from user provided settings.

    Acts as an entrypoint if user specifies --yaml flag in `torchlambda scheme`
    command.

    Parameters
    ----------
    args : dict-like
        User provided arguments parsed by argparse.ArgumentParser instance.

    """
    settings = read_settings(args)
    validator = utils.template.validator.get()
    validate(validator, settings)
    settings = validator.normalized(settings)

    source = create_source(settings)
    save(source, args)



import pathlib

import setuptools


def read(filepath, variable):
    namespace = {}
    exec(open(filepath).read(), namespace)
    return namespace[variable]


setuptools.setup(
    name="torchlambda",
    version=read("torchlambda/_version.py", "__version__"),
    license="MIT",
    author="Szymon Maszke",
    author_email="szymon.maszke@protonmail.com",
    description="Minimalistic & easy deployment of PyTorch models on AWS Lambda with C++",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/torchlambda",
    packages=setuptools.find_packages(),
    package_data={"": ["*.sh", "*.txt", "Dockerfile", ".dockerignore", "templates"]},
    python_requires=">=3.6",
    install_requires=["PyYAML>=5.3", "Cerberus>=1.3.2"],
    entry_points={"console_scripts": ["torchlambda=torchlambda.main:main"],},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    project_urls={
        "Website": "https://szymonmaszke.github.io/torchlambda",
        "Documentation": "https://szymonmaszke.github.io/torchlambda/#torchlambda",
        "Issues": "https://github.com/szymonmaszke/torchlambda/issues?q=is%3Aissue+is%3Aopen+sort%3Aupdated-desc",
    },
    keywords="aws lambda pytorch deployment minimal c++ neural network model",
)



import argparse


def settings(subparsers) -> None:
    """Create YAML settings to use with `torchlambda scheme --yaml`."""
    parser = subparsers.add_parser(
        "settings",
        help="Create YAML settings to use with `torchlambda scheme --yaml`.\n"
        "This is the easiest way to deploy model, just modify default settings provided by this comand.\n"
        "All unspecified fields will have default values just like those settings.\n",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--destination",
        required=False,
        default="./torchlambda.yaml",
        help="""Path to file where YAML settings will be stored. Default: "./torchlambda.yaml" """,
    )


def template(subparsers) -> None:
    """Create C++ source code template used for model inference."""
    parser = subparsers.add_parser(
        "template",
        help="Create C++ deployment code scheme with AWS Lambda C++ SDK and PyTorch.\n"
        "See generated code comments and change a few lines to fit your use case.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--yaml",
        required=False,
        default=None,
        help="Path to YAML containing settings to generate code scheme.\n"
        "See torchlambda settings and comments in generated YAML for more info.",
    )

    parser.add_argument(
        "--destination",
        required=False,
        default="./torchlambda",
        help="""Path to folder where C++ deployment files will be located. Default: "./torchlambda" """,
    )


def build(subparsers) -> None:
    """Perform deployment of PyTorch C++ code to AWS Lambda."""
    parser = subparsers.add_parser(
        "build",
        help="Obtain AWS Lambda ready .zip package from C++ deployment source code and Torchscript compiled model.\n"
        "See following resources for more information:\n"
        "- Torchscript documentation: https://pytorch.org/docs/stable/jit.html\n"
        "- AWS Lambda Getting Started: https://aws.amazon.com/lambda/getting-started/\n"
        "- AWS SDK for C++: https://aws.amazon.com/sdk-for-cpp/",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "source",
        help="Directory containing source files (possibly multiple) with C++ deployment code.\n"
        "cmake will build all files in this folder matching *.cpp, *.c, *.h or *.hpp extensions.\n"
        "If you are unsure how to create one, please use torchlambda scheme command.",
    )

    parser.add_argument(
        "--destination",
        default="./torchlambda.zip",
        required=False,
        help="Path where created AWS Lambda deployment .zip package will be stored.\n"
        "Should be a filename ending with .zip extension",
    )

    parser.add_argument(
        "--compilation",
        required=False,
        help="""Compilation flags used for inference source code.\n"""
        """Should be provided as string, e.g. "-Wall -Werror -O3".\n"""
        "By default no flags are passed to CMake targets.\n"
        'If you want to pass a single flag you should add space after string, e.g. "-Wall "\n'
        "Command line defaults pass none additional flags as well.\n"
        "User might want to specify -Os for smaller size or -O2 for possibly increased performance.\n"
        "IMPORTANT: Due to linker flags smaller binary size may not be possible at the moment.",
    )

    parser.add_argument(
        "--operations",
        required=False,
        help="Path containing exported model operations in .yaml format.\n"
        "See: https://pytorch.org/mobile/ios/#custom-build for more information.\n"
        "If specified, custom image will be build from scratch.\n"
        "Default: None (no operations)",
    )

    parser.add_argument(
        "--pytorch",
        nargs="+",
        required=False,
        default=[],
        help="PyTorch's libtorch build flags.\n"
        "See PyTorch's CMakeLists.txt for all available flags: https://github.com/pytorch/pytorch/blob/master/CMakeLists.txt\n"
        "If specified, custom image will be build from scratch.\n"
        "Default build parameters used:\n"
        "-DBUILD_PYTHON=OFF\n"
        "-DUSE_MPI=OFF\n"
        "-DUSE_NUMPY=OFF\n"
        "-DUSE_ROCM=OFF\n"
        "-DUSE_NCCL=OFF\n"
        "-DUSE_NUMA=OFF\n"
        "-DUSE_MKLDNN=OFF\n"
        "-DUSE_GLOO=OFF\n"
        "-DUSE_OPENMP=OFF\n"
        "User can override defaults by providing multiple arguments WITHOUT -D, e.g. \n"
        "--pytorch USE_NUMPY=ON USE_OPENMP=ON\n"
        "Default additional command line options: None",
    )

    parser.add_argument(
        "--pytorch-version",
        required=False,
        default=None,
        help="Commit or tag to which PyTorch will be set during build.\n"
        "See available releases at: https://github.com/pytorch/pytorch/releases (but any commit can be used)\n"
        'Special value "None" allowed which leaves PyTorch at current head on master.\n'
        "If specified, custom image will be build from scratch.\n"
        "Default: latest ",
    )

    parser.add_argument(
        "--aws",
        nargs="+",
        required=False,
        default=[],
        help="AWS C++ SDK build flags customizing dependency build.\n"
        "See: https://docs.aws.amazon.com/sdk-for-cpp/v1/developer-guide/cmake-params.html#cmake-build-only for more information.\n"
        "If specified, custom image will be built from scratch.\n"
        "Default build parameters used:\n"
        "-DBUILD_SHARED_LIBS=OFF (cannot be overriden)\n"
        "-DENABLE_UNITY_BUILD=ON (usually shouldn't be overriden)\n"
        "-DCUSTOM_MEMORY_MANAGEMENT=OFF\n"
        "-DCPP_STANDARD=17\n"
        "User can override defaults by providing multiple arguments WITHOUT -D, e.g. \n"
        "--aws CPP_STANDARD=11 CUSTOM_MEMORY_MANAGEMENT=ON\n"
        "`-DBUILD_ONLY` flag SHOULD NOT BE USED HERE, specify --aws-components instead\n"
        "Default additional command line options: None",
    )

    parser.add_argument(
        "--aws-components",
        nargs="+",
        default=[],
        required=False,
        help="Components of AWS C++ SDK to build.\n"
        "If specified, custom image will be built from scratch.\n"
        "Acts as `-DBUILD_ONLY` during build, please see https://docs.aws.amazon.com/sdk-for-cpp/v1/developer-guide/cmake-params.html#cmake-build-only.\n"
        "Pass components as space separated arguments, e.g.\n"
        "--aws-components s3 dynamodb\n"
        "By default only core will be build."
        "Default additional command line options: None",
    )

    parser.add_argument(
        "--image",
        required=False,
        default="torchlambda:custom",
        help="Name of Docker image to use for code building.\n"
        "If provided name image exists on localhost it will be used for `docker run` command.\n"
        "Otherwise AND IF custom build specified (either of --operations, --pytorch, --aws, --components or --build) image will be build from scratch.\n"
        "Otherwise prebuilt image szymonmaszke/torchlambda:latest will be downloaded and used.\n"
        "Default: torchlambda:custom",
    )

    parser.add_argument(
        "--docker",
        required=False,
        help='Flags passed to "docker" command during build and run.\n'
        'If you want to pass a single flag you should add space after string, e.g. "--debug "\n'
        """Flags should be passed as space separated string, e.g. "--debug --log-level debug".\nDefault: None""",
    )

    parser.add_argument(
        "--docker-build",
        required=False,
        help="Flags passed to docker build command (custom image building).\n"
        """Flags should be passed as space separated string, e.g. "--compress --no-cache".\n"""
        'If you want to pass a single flag you should add space after string, e.g. "--compress "\n'
        "`-t` flag SHOULD NOT BE USED, specify --image instead.\n"
        "Default: None",
    )

    parser.add_argument(
        "--docker-run",
        required=False,
        help="Flags passed to docker run command (code deployment building).\n"
        "Flags should be passed as space separated string, e.g.\n"
        """--name deployment --mount source=myvol2,target=/home/app".\n"""
        'If you want to pass a single flag you should add space after string, e.g. "--name my_name"\n'
        "Default: None",
    )

    parser.add_argument(
        "--no-run",
        required=False,
        action="store_true",
        help="Do not run compilation of source code part.\n"
        "Can be used to only create Docker building image to be run later."
        "Default: False",
    )


def layer(subparsers) -> None:
    """Pack model as .zip file ready to deploy on AWS Lambda as layer."""
    parser = subparsers.add_parser(
        "layer",
        help="Pack model as .zip file ready to deploy on AWS Lambda as layer.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "source",
        help="Path pointing to TorchScript compiled model.\n"
        "For more information check introduction to TorchScript:\n"
        "https://pytorch.org/tutorials/beginner/Intro_to_TorchScript_tutorial.html.",
    )

    parser.add_argument(
        "--destination",
        required=False,
        default="./model.zip",
        help="Path where AWS Lambda layer containing model will be stored.\n"
        """Default: "./model.zip" """,
    )

    parser.add_argument(
        "--directory",
        required=False,
        default=None,
        help="Directory where model will be stored. Usually you don't want to change that.\n"
        "Model will be unpacked to /opt/your/specified/directory and needs to be set analogously in C++/YAML settings.\n"
        "Default: None (model will be placed in /opt)",
    )

    parser.add_argument(
        "--compression",
        required=False,
        default="STORED",
        choices=["STORED", "DEFLATED", "BZIP2", "LZMA"],
        type=str.upper,
        help="""Compression method used for model compression.\n"""
        "See: https://docs.python.org/3/library/zipfile.html#zipfile.ZIP_STORED for more information.\n"
        "IMPORTANT: It's best to use default (uncompressed archive stored in.zip).\n"
        "Model .ptc file will barely be compressed by any algorithm, while it may increase decompression speed on AWS Lambda.\n"
        "If you wish to compress your model please use quantization (https://pytorch.org/docs/stable/quantization.html) or related techniques instead.\n"
        """Default: "STORED" """,
    )

    parser.add_argument(
        "--compression-level",
        required=False,
        default=None,
        choices=list(range(10)),
        type=int,
        help="""Level of compression used.\n"""
        "See: https://docs.python.org/3/library/zipfile.html#zipfile-objects for more information.\n"
        """Default: None (default for specified --compression) """,
    )
